<?php

namespace Builder\App;

use IoC\Container;
use Builder\App\Application\Bootstrap\Helpers as h;
use Builder\App\Providers\Provider;

class Application {

    const INI_FILE = '.env.ini';

    /**
     * @var
     */
    protected $container;

    /**
     * @var
     */
    protected $template;

    /**
     * @var
     */
    protected static $_instance;

    /**
     * @var
     */
    protected $config;

    /**
     * @var
     */
    protected $env;

    public function __construct() {
        $this->setContainer(
            new Container
        );

        $this->setConfig(
            Provider::load('app.php')
        );

        $file = dirname(__DIR__) . '/' . self::INI_FILE;

        if( ! file_exists($file) )
            throw new \Exception('Invalid ini file!');

        $env = parse_ini_string(
            file_get_contents($file)
        );

        $this->setEnv($env);
    }

    /**
     * @return mixed
     */
    public function getEnv() {
        return $this->env;
    }

    /**
     * @param mixed $env
     */
    public function setEnv($env) {
        $this->env = $env;
    }


    /**
     * Get config
     *
     * @param null $key
     * @param null $default
     * @return mixed
     */
    public function getConfig($key = null, $default = null) {
        return !is_null($key) ? $this->config->get($key, $default) : $this->config;
    }

    /**
     * @param mixed $config
     */
    public function setConfig($config) {
        $this->config = $config;
    }


    /**
     * Get instance .
     *
     * @return Application
     */
    public static function getInstance() {
        if ( ! ( self::$_instance instanceof self ) ) {
            self::$_instance = new self();
        }

        return self::$_instance;
    }


    /**
     * @return Container
     */
    public function getContainer() {
        return $this->container;
    }

    /**
     * @param Container $container
     */
    public function setContainer($container) {
        $this->container = $container;
    }


    /**
     * Register default providers .
     *
     */
    public function register() {
        $apps = $this;

        $config = $this->getConfig();

        $providers = is_admin() ? $config->get('providers.admin', []) : $config->get('providers.front', []);

        array_walk($providers, function($provider) use($apps) {
            $provider = new $provider($apps);
            $provider->register();

            /** Each provider can register boot method using wordPress registered hooks  */
            $provider->boot();
        });

        return $this;
    }


    /**
     * Activate plugin .
     * @param callable $closure
     * @return $this
     */
    public function on(\Closure $closure) {
        register_activation_hook( __FILE__, $closure );

        return $this;
    }

    /**
     * Deactivate plugin .
     * @param callable $closure
     * @return $this
     */
    public function off(\Closure $closure) {
        register_deactivation_hook( __FILE__, $closure );

        return $this;
    }

    /**
     * Call container directly .
     *
     * @param $name
     * @param array $arguments
     * @return mixed
     */
    public function __call($name, array $arguments) {
        $c = $this->getContainer();

        return call_user_func_array(array($c, $name), $arguments);
    }
}
